#lang racket/base

(require
  racket/cmdline
  racket/file
  racket/function
  racket/format
  racket/list
  racket/string
  "sudoku-solver/graph.rkt"
  "sudoku-solver/progress-data.rkt"
  "sudoku-solver/solver.rkt")

; String String -> String
;
; Add color to the output string where it differs from the input string.
(define (colorize-output input-string output-string)
  (define ESC "\u001b")
  (define COLOR-ON (format "~a[32m" ESC))
  (define COLOR-OFF (format "~a[0m" ESC))
  (string-join
    (for/list ([in-char input-string]
               [out-char output-string])
      (if (char=? in-char out-char)
          (~a out-char)
          (format "~a~a~a" COLOR-ON out-char COLOR-OFF)))
    ""))

; Integer -> String
;
; Return number string with thousands separators ("_") inserted.
(define (insert-thousands-separators number)
  (define (string-reverse string)
    (list->string (reverse (string->list string))))
  (string-trim
    (string-reverse
      (regexp-replace* #px"(\\d{3})"
                       (string-reverse (~a number))
                       "\\1_"))
    "_"))

; Solver-Result -> string
;
; Return solver statistics as a string.
(define (stats-string solver-result)
  (define MAX-BAR-WIDTH 60)
  (define BAR-DRAWING-CHARACTER #\u25ac)
  (define stats (recursion-stats (solver-result-progress-data solver-result)))
  ; At the end of the vector strip all consecutive items that are zero. There's
  ; no need to display those items. In particular, if there are n digits on the
  ; board to begin with, the maximum recursion level can be at most (81 - n - 1).
  (define trimmed-stats
    (dropf-right (vector->list stats)
                 (lambda (item) (= item 0))))
  (define max-counter (apply max (vector->list stats)))
  (string-append
    "\n"
    (format "Solver runtime: ~a seconds\n"
            (~r
              (/ (solver-result-duration solver-result) 1000)
              #:notation 'positional
              #:precision '(= 3)))
    (format "Total number of needed digit guesses: ~a\n"
            (insert-thousands-separators (solver-result-guesses-count solver-result)))
    "\n"
    "Recursion level / times that level was entered from above:\n"
    "\n"
    (string-join
      (for/list ([index (in-naturals 1)]
                 [counter (cdr trimmed-stats)])
        (format "~a / ~a ~a"
                (~r index #:min-width 2)
                (~a (insert-thousands-separators counter) #:align 'right #:min-width 9)
                (make-string (round (* MAX-BAR-WIDTH (/ counter max-counter)))
                             BAR-DRAWING-CHARACTER)))
      "\n")
    "\n"
    (format "Total: ~a\n"
            (insert-thousands-separators (apply + (vector->list stats))))
))

; Read the file given on the command line, solve it and print the result.
;
; Errors from invalid input boards or non-existent solutions currently aren't
; handled. These errors cause a stacktrace.
(define (main)
  (define new-file-path #f)
  (define new-use-color? #f)
  (define new-stats? #f)
  (define new-text-file-path #f)
  (define new-dot-file-path #f)
  (command-line
    #:program "sudoku-solver"
    #:once-each
    [("--color") "Colorize values of the solution."
                 (set! new-use-color? #t)]
    [("--stats")
     ("Print solver statistics."
      "These consist of:"
      "- solver runtime"
      "- number of needed digit guesses"
      "- number of times the solver encountered each recursion level. When the"
      "  solver consecutively tries the digits 1 to 9 at one recursion level, the"
      "  count for that level is only increased by one, not by the number of"
      "  digits tried. The initial recursion level of the solver is 1.")
     (set! new-stats? #t)]
    [("--progress-text-file")
     text-file
     ("Create a text file of the solution process."
      "The format of the lines is \"board-index --guess--> next-board-index\"."
      "Board indices are numbered row-wise from 0 for the upper-left corner to 80"
      "for the lower-right corner of the board/grid. The line means that the"
      "solver could make a feasible guess `guess` at `board-index` that made it"
      "possible to continue with guesses at `next-board-index`. The lines are"
      "indented according to the recursion level.")
     (set! new-text-file-path text-file)]
    [("--progress-dot-file")
     dot-file
     ("Create a DOT file for processing by Graphviz's `dot` command."
      "The generated diagram shows the solution process. The interpretation of the"
      "graph is analoguous to the text output from `--progress-text-file` (see"
      "above).")
     (set! new-dot-file-path dot-file)]
    #:ps
     ""
     "For more information visit the project website,"
     "https://git.sr.ht/~sschwarzer/sudoku-solver ."
    #:args (file-path)
    (set! new-file-path file-path))
  (define input-string (file->string new-file-path))
  (define solver-result (solve-board (string->board input-string)))
  (define output-string (board->string (solver-result-board solver-result)))
  ; Normalize input string. The coloring algorithm doesn't work if the input
  ; and output string are out of sync, for example because the input string
  ; contains fewer or more newlines.
  (define normalized-input-string (board->string (string->board input-string)))
  (printf "Input:\n\n~a\n\nSolution:\n\n~a\n"
          normalized-input-string
          (if new-use-color?
              (colorize-output normalized-input-string output-string)
              output-string))
  (when new-stats?
    (printf (stats-string solver-result)))
  (when new-text-file-path
    (displayln "\nWriting solution text file ...")
    (with-output-to-file
      new-text-file-path
      (thunk
        (progress-data->text-file
          (solver-result-progress-data solver-result)
          (current-output-port)))
      #:mode 'text
      #:exists 'truncate))
  (when new-dot-file-path
    (displayln "\nWriting solution dot file ...")
    (with-output-to-file
      new-dot-file-path
      (thunk
        (node->dot-file
          (progress-data->node
            (solver-result-progress-data solver-result))
            (current-output-port)))
      #:mode 'text
      #:exists 'truncate)))

(error-display-handler
  (lambda (message exn)
    (eprintf (regexp-replace #px"^.*?: " message "error: "))))

(module+ main
  (main))
