#lang racket/base

(require racket/contract)

(provide
  (contract-out
    (struct node ([node-label string?]
                  [edge-label string?]
                  [child-nodes (listof node?)]))
    [node->dot-file (-> node? output-port? any)]
))

(require
  racket/generator
  racket/string)

; String String [List-of Node]
(struct node
  (; Label inside node shape
   node-label
   ; Label of edge _to_ this node. `#f` if no label
   edge-label
   ; List of nodes
   child-nodes)
  #:constructor-name Node)

; -> [Generator-of String]
;
; Return a generator that yields consecutive ids of the form "n<number>".
(define (make-node-id-generator)
  (generator ()
    (for ([i (in-naturals 1)])
      (yield (format "n~a" i)))))

; Node Output-port ->
;
; Write Graphviz graph representation to `output-port`.
(define (node->dot-file node output-port)
  (define node-ranks (make-hash))
  (define make-node-id (make-node-id-generator))
  (parameterize ([current-output-port output-port])
    (displayln "digraph {")
    (displayln "rankdir=LR")
    (let render-node ([node node]
                      [node-id (make-node-id)]
                      [depth 1])
      (hash-set! node-ranks
                 depth
                 (cons node-id (hash-ref node-ranks depth '())))
      (displayln (format "~a [label=\"~a\"]"
                         node-id
                         (node-node-label node)))
      ; Render child nodes
      (for ([child-node (node-child-nodes node)])
        (define child-node-id (make-node-id))
        ; Render sub-graph for child node.
        (render-node child-node child-node-id (add1 depth))
        ; Render connection to child node.
        (displayln (format "~a -> ~a [label=\" ~a \"]"
                           node-id
                           child-node-id
                           (node-edge-label child-node)))))
    ; Add rank information.
    (for ([depth (hash-keys node-ranks)])
      (displayln (format "{ rank=\"same\"; ~a }"
                         (string-join
                           (reverse (hash-ref node-ranks depth))
                           "; "))))
    (displayln "}")))

; --------------------------------------------------------------------------------

#|
(define example-graph
  (Node "foo"
        #f
        (list (Node "bar"
                    "foo_to_bar"
                    (list (Node "bar_child" "bar_to_bar_child" '())))
              (Node "baz" "foo_to_baz" '()))))

(render-graph example-graph (current-output-port))
|#
